import logging
import os
import plistlib

import click
import coloredlogs

NOTIFICATIONS_FILENAME = os.path.join(os.path.dirname(__file__), 'notifications.txt')

coloredlogs.install(level=logging.DEBUG)


def get_notifications():
    with open(NOTIFICATIONS_FILENAME, 'rb') as f:
        return f.read().decode().split('\n')


def save_notifications(notifications):
    with open(NOTIFICATIONS_FILENAME, 'wb') as f:
        f.write('\n'.join(notifications).encode())


@click.command()
@click.argument('root_fs', type=click.Path(dir_okay=True, file_okay=False, exists=True))
def main(root_fs):
    """
    Add notifications registered to `com.apple.notifyd.matching` from a given IPSW `root_fs` (extracted filesystem)
    into `notifications.txt`
    """
    launch_daemons = os.path.join(root_fs, 'System', 'Library', 'LaunchDaemons')

    notifications = set(get_notifications())

    for filename in os.listdir(launch_daemons):
        if not filename.endswith('.plist'):
            continue

        filename = os.path.join(launch_daemons, filename)
        with open(filename, 'rb') as f:
            plist = plistlib.load(f)

        launch_events = plist.get('LaunchEvents', {})
        notifyd_matching = launch_events.get('com.apple.notifyd.matching', {})

        for v in notifyd_matching.values():
            if not isinstance(v, dict):
                logging.error(f'error parsing: {filename}')
                continue
            notification = v.get('Notification')
            if notification is None:
                continue

            if notification not in notifications:
                logging.info(f'adding notification: {notification}')
                notifications.add(notification)

    save_notifications(notifications)


if __name__ == '__main__':
    main()



from typing import Mapping, List

from pymobiledevice3.exceptions import PyMobileDevice3Exception, NotMountedError, UnsupportedCommandError, \
    AlreadyMountedError, InternalError, DeveloperModeIsNotEnabledError, MessageNotSupportedError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService


class MobileImageMounterService(BaseService):
    SERVICE_NAME = 'com.apple.mobile.mobile_image_mounter'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def copy_devices(self) -> List[Mapping]:
        """ Copy mounted devices list. """
        try:
            return self.service.send_recv_plist({'Command': 'CopyDevices'})['EntryList']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def lookup_image(self, image_type: str) -> bytes:
        """ Lookup mounted image by its name. """
        response = self.service.send_recv_plist({'Command': 'LookupImage',
                                                 'ImageType': image_type})

        if not response or not response.get('ImagePresent', True):
            raise NotMountedError()

        signature = response.get('ImageSignature', [])
        if isinstance(signature, list):
            if not signature:
                raise NotMountedError()
            return signature[0]
        return signature

    def is_image_mounted(self, image_type: str) -> bool:
        try:
            self.lookup_image(image_type)
            return True
        except NotMountedError:
            return False

    def umount(self, mount_path: str, image_type: str = None, signature: bytes = None) -> None:
        """ umount image. """
        request = {'Command': 'UnmountImage', 'MountPath': mount_path}

        if image_type is not None:
            request['ImageType'] = image_type

        if signature is not None:
            request['ImageSignature'] = signature

        response = self.service.send_recv_plist(request)

        error = response.get('Error')
        if error:
            if error == 'UnknownCommand':
                raise UnsupportedCommandError()
            elif error == 'InternalError':
                raise InternalError()
            else:
                raise NotMountedError()

    def mount(self, image_type: str, signature: bytes, trust_cache: bytes = None, info_plist: Mapping = None) -> None:
        """ Upload image into device. """

        if self.is_image_mounted(image_type):
            raise AlreadyMountedError()

        request = {'Command': 'MountImage',
                   'ImageType': image_type,
                   'ImageSignature': signature}

        if image_type == 'Cryptex':
            if trust_cache is None:
                raise ValueError('Cryptex image requires a ImageTrustCache to be also supplied')

            if info_plist is None:
                raise ValueError('Cryptex image requires a ImageInfoPlist to be also supplied')

            request['ImageTrustCache'] = trust_cache
            request['ImageInfoPlist'] = info_plist

        response = self.service.send_recv_plist(request)

        if 'Developer mode is not enabled' in response.get('DetailedError', ''):
            raise DeveloperModeIsNotEnabledError()

        status = response.get('Status')

        if status != 'Complete':
            raise PyMobileDevice3Exception(f'command MountImage failed with: {response}')

    def upload_image(self, image_type: str, image: bytes, signature: bytes) -> None:
        """ Upload image into device. """
        self.service.send_plist({'Command': 'ReceiveBytes',
                                 'ImageType': image_type,
                                 'ImageSize': len(image),
                                 'ImageSignature': signature})
        result = self.service.recv_plist()

        status = result.get('Status')

        if status != 'ReceiveBytesAck':
            raise PyMobileDevice3Exception(f'command ReceiveBytes failed with: {result}')

        self.service.sendall(image)
        result = self.service.recv_plist()

        status = result.get('Status')

        if status != 'Complete':
            raise PyMobileDevice3Exception(f'command ReceiveBytes failed to send bytes with: {result}')

    def query_developer_mode_status(self) -> bool:
        response = self.service.send_recv_plist({'Command': 'QueryDeveloperModeStatus'})

        try:
            return response['DeveloperModeStatus']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def query_nonce(self) -> bytes:
        response = self.service.send_recv_plist({'Command': 'QueryNonce'})

        try:
            return response['PersonalizationNonce']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def query_personalization_identifiers(self, image_type: str = None) -> bytes:
        request = {'Command': 'QueryPersonalizationIdentifiers'}

        if image_type is not None:
            request['PersonalizedImageType'] = image_type

        response = self.service.send_recv_plist(request)

        try:
            return response['PersonalizationIdentifiers']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def query_personalization_manifest(self, image_type: str, signature: bytes) -> Mapping:
        request = {'Command': 'QueryPersonalizationManifest', 'PersonalizedImageType': image_type,
                   'ImageType': image_type, 'ImageSignature': signature}

        response = self.service.send_recv_plist(request)

        try:
            # The response is returned as "ImageSignature" which is wrong, but that's what Apple does
            return response['ImageSignature']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def roll_personalization_nonce(self) -> None:
        try:
            self.service.send_recv_plist({'Command': 'RollPersonalizationNonce'})
        except ConnectionAbortedError:
            return

    def roll_cryptex_nonce(self) -> None:
        try:
            self.service.send_recv_plist({'Command': 'RollCryptexNonce'})
        except ConnectionAbortedError:
            return



import logging

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.service_connection import ServiceConnection


class BaseService:
    def __init__(self, lockdown: LockdownClient, service_name: str, is_developer_service=False,
                 service: ServiceConnection = None):
        """
        :param lockdown: lockdown connection
        :param service_name: wrapped service name - will attempt
        :param is_developer_service: should DeveloperDiskImage be mounted before
        :param service: an established service connection object. If none, will attempt connecting to service_name
        """

        if not service:
            start_service = lockdown.start_developer_service if is_developer_service else lockdown.start_service
            service = start_service(service_name)

        self.service_name = service_name
        self.lockdown = lockdown
        self.service = service
        self.logger = logging.getLogger(self.__module__)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.service.close()



import asyncio
from functools import partial

from pymobiledevice3.exceptions import WirError


class SessionProtocol:
    def __init__(self, inspector, id_, app, page, method_prefix='Automation'):
        """
        :param pymobiledevice3.services.webinspector.WebinspectorService inspector:
        """
        self.app = app
        self.page = page
        self.inspector = inspector
        self.id_ = id_
        self.method_prefix = method_prefix
        self._wir_messages_id = 1

    async def send_command(self, method, **kwargs):
        wir_id = self._wir_messages_id
        self._wir_messages_id += 1
        await self.inspector.send_socket_data(self.id_, self.app.id_, self.page.id_, {
            'method': f'{self.method_prefix}.{method}' if self.method_prefix else method,
            'params': kwargs,
            'id': wir_id,
        })
        return wir_id

    async def get_response(self, wir_id):
        response = await self.wait_for_message(wir_id)
        if 'result' in response:
            return response['result']
        elif 'error' in response:
            raise WirError(response['error']['message'])

    async def send_receive(self, method, wait_for_response=True, **kwargs):
        wir_id = await self.send_command(method, **kwargs)
        if wait_for_response:
            return await self.get_response(wir_id)
        else:
            return wir_id

    async def wait_for_message(self, id_):
        while id_ not in self.inspector.wir_message_results:
            await asyncio.sleep(0)
        return self.inspector.wir_message_results.pop(id_)

    def sync_send_receive(self, method, wait_for_response=True, **kwargs):
        return self.inspector.await_(self.send_receive(method, wait_for_response, **kwargs))

    def __getattr__(self, item):
        return partial(self.sync_send_receive, method=item)



import logging

import pyimg4
from pyimg4 import IM4P

logger = logging.getLogger(__name__)

COMPONENT_FOURCC = {
    'ACIBT': 'acib',
    'ACIBTLPEM': 'lpbt',
    'ACIWIFI': 'aciw',
    'ANE': 'anef',
    'ANS': 'ansf',
    'AOP': 'aopf',
    'AVE': 'avef',
    'Alamo': 'almo',
    'Ap,ANE1': 'ane1',
    'Ap,ANE2': 'ane2',
    'Ap,ANE3': 'ane3',
    'Ap,AudioAccessibilityBootChime': 'auac',
    'Ap,AudioBootChime': 'aubt',
    'Ap,AudioPowerAttachChime': 'aupr',
    'Ap,BootabilityBrainTrustCache': 'trbb',
    'Ap,CIO': 'ciof',
    'Ap,HapticAssets': 'hpas',
    'Ap,LocalBoot': 'lobo',
    'Ap,LocalPolicy': 'lpol',
    'Ap,NextStageIM4MHash': 'nsih',
    'Ap,RecoveryOSPolicyNonceHash': 'ronh',
    'Ap,RestoreANE1': 'ran1',
    'Ap,RestoreANE2': 'ran2',
    'Ap,RestoreANE3': 'ran3',
    'Ap,RestoreCIO': 'rcio',
    'Ap,RestoreTMU': 'rtmu',
    'Ap,Scorpius': 'scpf',
    'Ap,SystemVolumeCanonicalMetadata': 'msys',
    'Ap,TMU': 'tmuf',
    'Ap,VolumeUUID': 'vuid',
    'Ap,rOSLogo1': 'rlg1',
    'Ap,rOSLogo2': 'rlg2',
    'AppleLogo': 'logo',
    'AudioCodecFirmware': 'acfw',
    'BatteryCharging': 'glyC',
    'BatteryCharging0': 'chg0',
    'BatteryCharging1': 'chg1',
    'BatteryFull': 'batF',
    'BatteryLow0': 'bat0',
    'BatteryLow1': 'bat1',
    'BatteryPlugin': 'glyP',
    'CFELoader': 'cfel',
    'CrownFirmware': 'crwn',
    'DCP': 'dcpf',
    'Dali': 'dali',
    'DeviceTree': 'dtre',
    'Diags': 'diag',
    'EngineeringTrustCache': 'dtrs',
    'ExtDCP': 'edcp',
    'GFX': 'gfxf',
    'Hamm': 'hamf',
    'Homer': 'homr',
    'ISP': 'ispf',
    'InputDevice': 'ipdf',
    'KernelCache': 'krnl',
    'LLB': 'illb',
    'LeapHaptics': 'lphp',
    'Liquid': 'liqd',
    'LoadableTrustCache': 'ltrs',
    'LowPowerWallet0': 'lpw0',
    'LowPowerWallet1': 'lpw1',
    'LowPowerWallet2': 'lpw2',
    'MacEFI': 'mefi',
    'MtpFirmware': 'mtpf',
    'Multitouch': 'mtfw',
    'NeedService': 'nsrv',
    'OS': 'OS\0\0',
    'OSRamdisk': 'osrd',
    'PEHammer': 'hmmr',
    'PERTOS': 'pert',
    'PHLEET': 'phlt',
    'PMP': 'pmpf',
    'PersonalizedDMG': 'pdmg',
    'RBM': 'rmbt',
    'RTP': 'rtpf',
    'Rap,SoftwareBinaryDsp1': 'sbd1',
    'Rap,RTKitOS': 'rkos',
    'Rap,RestoreRTKitOS': 'rrko',
    'RecoveryMode': 'recm',
    'RestoreANS': 'rans',
    'RestoreDCP': 'rdcp',
    'RestoreDeviceTree': 'rdtr',
    'RestoreExtDCP': 'recp',
    'RestoreKernelCache': 'rkrn',
    'RestoreLogo': 'rlgo',
    'RestoreRTP': 'rrtp',
    'RestoreRamDisk': 'rdsk',
    'RestoreSEP': 'rsep',
    'RestoreTrustCache': 'rtsc',
    'SCE': 'scef',
    'SCE1Firmware': 'sc1f',
    'SEP': 'sepi',
    'SIO': 'siof',
    'StaticTrustCache': 'trst',
    'SystemLocker': 'lckr',
    'SystemVolume': 'isys',
    'WCHFirmwareUpdater': 'wchf',
    'ftap': 'ftap',
    'ftsp': 'ftsp',
    'iBEC': 'ibec',
    'iBSS': 'ibss',
    'iBoot': 'ibot',
    'iBootData': 'ibdt',
    'iBootDataStage1': 'ibd1',
    'iBootTest': 'itst',
    'rfta': 'rfta',
    'rfts': 'rfts',
    'Ap,DCP2': 'dcp2',
}


def stitch_component(name: str, im4p_data: bytes, tss):
    logger.info(f'Personalizing IMG4 component {name}...')

    im4p = IM4P(data=im4p_data)

    # patch fourcc
    fourcc = COMPONENT_FOURCC.get(name)
    if fourcc is not None:
        im4p.fourcc = fourcc

    # hack if we have a *-TBM entry for the give
    tbm_dict = tss.get(f'{name}-TBM')

    im4r = None
    if tbm_dict is not None:
        im4r = pyimg4.IM4R()
        for key in tbm_dict.keys():
            im4r.add_property(pyimg4.RestoreProperty(fourcc=key, value=tbm_dict[key]))
    return pyimg4.IMG4(im4p=im4p, im4m=tss.ap_img4_ticket, im4r=im4r).output()



import logging
import plistlib
import tempfile
import zipfile
from pathlib import Path
from typing import IO

import click
import requests
from tqdm import tqdm

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.exceptions import NotMountedError, UnsupportedCommandError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.mobile_image_mounter import MobileImageMounterService

DEVELOPER_DISK_IMAGE_URL = 'https://github.com/pdso/DeveloperDiskImage/raw/master/{ios_version}/{ios_version}.zip'

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ mounter cli """
    pass


@cli.group()
def mounter():
    """ mounter options """
    pass


@mounter.command('list', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_list(lockdown: LockdownClient, color):
    """ list all mounted images """
    output = []

    images = MobileImageMounterService(lockdown=lockdown).copy_devices()
    for image in images:
        image_signature = image.get('ImageSignature')
        if image_signature is not None:
            image['ImageSignature'] = image_signature.hex()
        output.append(image)

    print_json(output, colored=color)


@mounter.command('lookup', cls=Command)
@click.option('--color/--no-color', default=True)
@click.argument('image_type')
def mounter_lookup(lockdown: LockdownClient, color, image_type):
    """ lookup mounter image type """
    try:
        signature = MobileImageMounterService(lockdown=lockdown).lookup_image(image_type)
        print_json(signature, colored=color)
    except NotMountedError:
        logger.error(f'Disk image of type: {image_type} is not mounted')


@mounter.command('umount', cls=Command)
@click.option('-t', '--image-type', type=click.Choice(['Developer', 'Cryptex']), default='Developer')
@click.option('-p', '--mount-path', help='Only needed for older iOS version', default='/Developer')
def mounter_umount(lockdown: LockdownClient, image_type: str, mount_path: str):
    """ unmount developer image. """
    image_mounter = MobileImageMounterService(lockdown=lockdown)
    try:
        image_mounter.umount(mount_path, image_type=image_type, signature=b'')
        logger.info('DeveloperDiskImage unmounted successfully')
    except NotMountedError:
        logger.error('DeveloperDiskImage isn\'t currently mounted')
    except UnsupportedCommandError:
        logger.error('Your iOS version doesn\'t support this command')


def download_file(url, local_filename):
    logger.debug(f'downloading: {local_filename}')
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        total_size_in_bytes = int(r.headers.get('content-length', 0))

        with tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True, dynamic_ncols=True) as progress_bar:
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    progress_bar.update(len(chunk))
                    f.write(chunk)

    return local_filename


def download_developer_disk_image(ios_version, directory):
    url = DEVELOPER_DISK_IMAGE_URL.format(ios_version=ios_version)
    with tempfile.NamedTemporaryFile('wb+') as f:
        download_file(url, f.name)
        zip_file = zipfile.ZipFile(f)
        zip_file.extractall(directory)


@mounter.command('mount', cls=Command)
@click.argument('image-path', type=click.Path(exists=True))
@click.argument('signature', type=click.Path(exists=True))
@click.argument('image-type', type=click.Choice(['Developer', 'Cryptex']), default='Developer')
@click.option('--trust-cache', type=click.File('rb'), help='Used only for Cryptex images')
@click.option('--info-plist', type=click.File('rb'), help='Used only for Cryptex images')
def mounter_mount(lockdown: LockdownClient, image_path: str, signature: str, image_type: str, trust_cache: IO = None,
                  info_plist: IO = None):
    """ mount developer image. """
    image_mounter = MobileImageMounterService(lockdown=lockdown)
    if image_mounter.is_image_mounted(image_type):
        logger.error(f'{image_type} is already mounted')
        return

    if trust_cache is not None:
        trust_cache = trust_cache.read()

    if info_plist is not None:
        info_plist = plistlib.load(info_plist)

    image_path = Path(image_path)
    signature = Path(signature)
    image_path = image_path.read_bytes()
    signature = signature.read_bytes()

    image_mounter.upload_image(image_type, image_path, signature)
    image_mounter.mount(image_type, signature, trust_cache=trust_cache, info_plist=info_plist)
    logger.info(f'{image_type} mounted successfully')


@mounter.command('auto-mount', cls=Command)
@click.option('-x', '--xcode', type=click.Path(exists=True, dir_okay=True, file_okay=False),
              help='Xcode application path used to figure out automatically the DeveloperDiskImage path')
@click.option('-v', '--version', help='use a different DeveloperDiskImage version from the one retrieved by lockdown'
                                      'connection')
def mounter_auto_mount(lockdown: LockdownClient, xcode: str, version: str):
    """ auto-detect correct DeveloperDiskImage and mount it """
    image_type = 'Developer'

    if xcode is None:
        # avoid "default"-ing this option, because Windows and Linux won't have this path
        xcode = Path('/Applications/Xcode.app')
        if not (xcode.exists()):
            xcode = Path('~/.pymobiledevice3/Xcode.app').expanduser()
            xcode.mkdir(parents=True, exist_ok=True)

    image_mounter = MobileImageMounterService(lockdown=lockdown)
    if image_mounter.is_image_mounted(image_type):
        logger.error('DeveloperDiskImage is already mounted')
        return

    logger.debug('trying to figure out the best suited DeveloperDiskImage')
    if version is None:
        version = lockdown.sanitized_ios_version
    image_dir = f'{xcode}/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport/{version}'
    image_path = f'{image_dir}/DeveloperDiskImage.dmg'
    signature = f'{image_path}.signature'
    developer_disk_image_dir = Path(image_path).parent

    if not developer_disk_image_dir.exists():
        try:
            download_developer_disk_image(version, developer_disk_image_dir)
        except PermissionError:
            logger.error(
                f'DeveloperDiskImage could not be saved to Xcode default path ({developer_disk_image_dir}). '
                f'Please make sure your user has the necessary permissions')
            return

    image_path = Path(image_path)
    signature = Path(signature)
    image_path = image_path.read_bytes()
    signature = signature.read_bytes()

    image_mounter.upload_image(image_type, image_path, signature)
    image_mounter.mount(image_type, signature)
    logger.info('DeveloperDiskImage mounted successfully')


@mounter.command('query-developer-mode-status', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_developer_mode_status(lockdown: LockdownClient, color):
    """ Query developer mode status """
    print_json(MobileImageMounterService(lockdown=lockdown).query_developer_mode_status(), colored=color)


@mounter.command('query-nonce', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_nonce(lockdown: LockdownClient, color):
    """ Query nonce """
    print_json(MobileImageMounterService(lockdown=lockdown).query_nonce(), colored=color)


@mounter.command('query-personalization-identifiers', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_personalization_identifiers(lockdown: LockdownClient, color):
    """ Query personalization identifiers """
    print_json(MobileImageMounterService(lockdown=lockdown).query_personalization_identifiers(), colored=color)


@mounter.command('roll-personalization-nonce', cls=Command)
def mounter_roll_personalization_nonce(lockdown: LockdownClient):
    MobileImageMounterService(lockdown=lockdown).roll_personalization_nonce()


@mounter.command('roll-cryptex-nonce', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_roll_cryptex_nonce(lockdown: LockdownClient):
    """ Roll cryptex nonce (will reboot) """
    MobileImageMounterService(lockdown=lockdown).roll_cryptex_nonce()



import dataclasses
import ipaddress
import logging

from construct import Struct, Int32ul, this, Adapter, Switch, Int8ul, Int16ub, Bytes


class IpAddressAdapter(Adapter):
    def _decode(self, obj, context, path):
        return ipaddress.ip_address(obj)


address_t = Struct(
    'len' / Int8ul,
    'family' / Int8ul,
    'port' / Int16ub,
    'data' / Switch(this.len, {
        0x1c: Struct(
            'flow_info' / Int32ul,
            'address' / IpAddressAdapter(Bytes(16)),
            'scope_id' / Int32ul,
        ),
        0x10: Struct(
            'address' / IpAddressAdapter(Bytes(4)),
            '_zero' / Bytes(8)
        )
    })

)

MESSAGE_TYPE_INTERFACE_DETECTION = 0
MESSAGE_TYPE_CONNECTION_DETECTION = 1
MESSAGE_TYPE_CONNECTION_UPDATE = 2


@dataclasses.dataclass
class InterfaceDetectionEvent:
    interface_index: int
    name: str


@dataclasses.dataclass
class ConnectionDetectionEvent:
    local_address: str
    remote_address: str
    interface_index: int
    pid: int
    recv_buffer_size: int
    recv_buffer_used: int
    serial_number: int
    kind: int


@dataclasses.dataclass
class ConnectionUpdateEvent:
    rx_packets: int
    rx_bytes: int
    tx_bytes: int
    rx_dups: int
    rx000: int
    tx_retx: int
    min_rtt: int
    avg_rtt: int
    connection_serial: int
    unknown0: int
    unknown1: int


class NetworkMonitor:
    IDENTIFIER = 'com.apple.instruments.server.services.networking'

    def __init__(self, dvt):
        self.logger = logging.getLogger(__name__)
        self._channel = dvt.make_channel(self.IDENTIFIER)

    def __enter__(self):
        self._channel.startMonitoring(expects_reply=False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._channel.stopMonitoring()

    def __iter__(self):
        while True:
            message = self._channel.receive_plist()

            event = None

            if message is None:
                continue

            if message[0] == MESSAGE_TYPE_INTERFACE_DETECTION:
                event = InterfaceDetectionEvent(*message[1])
            elif message[0] == MESSAGE_TYPE_CONNECTION_DETECTION:
                event = ConnectionDetectionEvent(*message[1])
                event.local_address = address_t.parse(event.local_address)
                event.remote_address = address_t.parse(event.remote_address)
            elif message[0] == MESSAGE_TYPE_CONNECTION_UPDATE:
                event = ConnectionUpdateEvent(*message[1])
            else:
                self.logger.warning(f'unsupported event type: {message[0]}')
            yield event



from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService
from pymobiledevice3.utils import try_decode

CHUNK_SIZE = 4096
TIME_FORMAT = '%H:%M:%S'
SYSLOG_LINE_SPLITTER = b'\n\x00'


class SyslogService(BaseService):
    """
    View system logs
    """

    SERVICE_NAME = 'com.apple.syslog_relay'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def watch(self):
        buf = b''
        while True:
            # read in chunks till we have at least one syslog line
            chunk = self.service.recv(CHUNK_SIZE)
            buf += chunk

            # SYSLOG_LINE_SPLITTER is used to split each syslog line
            if SYSLOG_LINE_SPLITTER in buf:
                lines = buf.split(SYSLOG_LINE_SPLITTER)

                # handle partial last lines
                if not buf.endswith(SYSLOG_LINE_SPLITTER):
                    buf = lines[-1]
                    lines = lines[:-1]

                for line in lines:
                    if len(line) == 0:
                        continue

                    yield try_decode(line)



import os
import posixpath

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import AfcService
from pymobiledevice3.services.base_service import BaseService

client_options = {
    'SkipUninstall': False,
    'ApplicationSINF': False,
    'iTunesMetadata': False,
    'ReturnAttributes': False
}


class InstallationProxyService(BaseService):
    SERVICE_NAME = 'com.apple.mobile.installation_proxy'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def watch_completion(self, handler=None, *args):
        while True:
            response = self.service.recv_plist()
            if not response:
                break
            error = response.get('Error')
            if error:
                raise IOError(f'''{error}: {response.get('ErrorDescription')}''')
            completion = response.get('PercentComplete')
            if completion:
                if handler:
                    self.logger.debug('calling handler')
                    handler(completion, *args)
                self.logger.info('%s %% Complete', response.get('PercentComplete'))
            if response.get('Status') == 'Complete':
                return response.get('Status')
        return 'Error'

    def send_cmd_for_bid(self, bid, cmd='Archive', options=None, handler=None, *args):
        cmd = {'Command': cmd,
               'ApplicationIdentifier': bid}
        if options:
            cmd.update({'ClientOptions': options})
        self.service.send_plist(cmd)
        self.logger.info('%s : %s\n', cmd, self.watch_completion(handler, *args))

    def uninstall(self, bid, options=None, handler=None, *args):
        return self.send_cmd_for_bid(bid, 'Uninstall', options, handler, args)

    def install_from_local(self, ipa_path, cmd='Install', options=None, handler=None, *args):
        if options is None:
            options = {}
        remote_path = posixpath.join('/', os.path.basename(ipa_path))
        afc = AfcService(self.lockdown)
        afc.set_file_contents(remote_path, open(ipa_path, 'rb').read())
        cmd = {'Command': cmd,
               'ClientOptions': options,
               'PackagePath': remote_path}
        self.service.send_plist(cmd)
        self.watch_completion(handler, args)

    def install(self, ipa_path, options=None, handler=None, *args):
        if options is None:
            options = {}
        return self.install_from_local(ipa_path, 'Install', options, handler, args)

    def upgrade(self, ipa_path, options=None, handler=None, *args):
        if options is None:
            options = {}
        return self.install_from_local(ipa_path, 'Upgrade', options, handler, args)

    def check_capabilities_match(self, capabilities, options=None):
        if options is None:
            options = {}
        cmd = {'Command': 'CheckCapabilitiesMatch',
               'ClientOptions': options}

        if capabilities:
            cmd['Capabilities'] = capabilities

        self.service.send_plist(cmd)
        result = self.service.recv_plist().get('LookupResult')
        return result

    def browse(self, options=None, attributes=None):
        if options is None:
            options = {}
        if attributes:
            options['ReturnAttributes'] = attributes

        cmd = {'Command': 'Browse',
               'ClientOptions': options}

        self.service.send_plist(cmd)

        result = []
        while True:
            z = self.service.recv_plist()
            if not z:
                break

            data = z.get('CurrentList')
            if data:
                result += data

            if z.get('Status') == 'Complete':
                break

        return result

    def apps_info(self, options=None):
        if options is None:
            options = {}
        cmd = {'Command': 'Lookup',
               'ClientOptions': options}

        self.service.send_plist(cmd)
        return self.service.recv_plist().get('LookupResult')

    def archive(self, bid, options=None, handler=None, *args):
        if options is None:
            options = {}
        self.send_cmd_for_bid(bid, 'Archive', options, handler, args)

    def restore_archive(self, bid, options=None, handler=None, *args):
        if options is None:
            options = {}
        self.send_cmd_for_bid(bid, 'Restore', options, handler, args)

    def remove_archive(self, bid, options=None, handler=None, *args):
        if options is None:
            options = {}
        self.send_cmd_for_bid(bid, 'RemoveArchive', options, handler, args)

    def archives_info(self, options=None):
        if options is None:
            options = {}
        cmd = {'Command': 'LookupArchive',
               'ClientOptions': options}
        return self.service.send_recv_plist(cmd).get('LookupResult')

    def search_path_for_bid(self, bid):
        path = None
        for a in self.get_apps(app_types=['User', 'System']):
            if a.get('CFBundleIdentifier') == bid:
                path = a.get('Path') + '/' + a.get('CFBundleExecutable')
        return path

    def get_apps(self, app_types=None):
        if app_types is None:
            app_types = ['User']
        return [app for app in self.apps_info().values()
                if app.get('ApplicationType') in app_types]

    def print_apps(self, app_types=None):
        if app_types is None:
            app_types = ['User']
        for app in self.get_apps(app_types):
            print(('%s : %s => %s' % (app.get('CFBundleDisplayName'),
                                      app.get('CFBundleIdentifier'),
                                      app.get('Path') if app.get('Path')
                                      else app.get('Container'))).encode('utf-8'))

    def get_apps_bid(self, app_types=None):
        if app_types is None:
            app_types = ['User']
        return [app['CFBundleIdentifier']
                for app in self.get_apps()
                if app.get('ApplicationType') in app_types]

    def close(self):
        self.service.close()



from abc import abstractmethod, ABC
from base64 import b64decode

from pymobiledevice3.services.web_protocol.automation_session import By


class SeleniumApi(ABC):
    @abstractmethod
    def find_element(self, by=By.ID, value=None):
        pass

    @abstractmethod
    def find_elements(self, by=By.ID, value=None):
        pass

    @property
    @abstractmethod
    def screenshot_as_base64(self):
        pass

    def find_element_by_class_name(self, name):
        return self.find_element(By.CLASS_NAME, name)

    def find_element_by_css_selector(self, css_selector):
        return self.find_element(By.CSS_SELECTOR, css_selector)

    def find_element_by_id(self, id_):
        return self.find_element(value=id_)

    def find_element_by_link_text(self, link_text):
        return self.find_element(By.LINK_TEXT, link_text)

    def find_element_by_name(self, name):
        return self.find_element(By.NAME, name)

    def find_element_by_partial_link_text(self, link_text):
        return self.find_element(By.PARTIAL_LINK_TEXT, link_text)

    def find_element_by_tag_name(self, name):
        return self.find_element(By.TAG_NAME, name)

    def find_element_by_xpath(self, xpath):
        return self.find_element(By.XPATH, xpath)

    def find_elements_by_class_name(self, name):
        return self.find_elements(By.CLASS_NAME, name)

    def find_elements_by_css_selector(self, css_selector):
        return self.find_elements(By.CSS_SELECTOR, css_selector)

    def find_elements_by_id(self, id_):
        return self.find_elements(value=id_)

    def find_elements_by_link_text(self, link_text):
        return self.find_elements(By.LINK_TEXT, link_text)

    def find_elements_by_name(self, name):
        return self.find_elements(By.NAME, name)

    def find_elements_by_partial_link_text(self, link_text):
        return self.find_elements(By.PARTIAL_LINK_TEXT, link_text)

    def find_elements_by_tag_name(self, name):
        return self.find_elements(By.TAG_NAME, name)

    def find_elements_by_xpath(self, xpath):
        return self.find_elements(By.XPATH, xpath)

    def screenshot(self, filename):
        png = self.screenshot_as_png()
        try:
            with open(filename, 'wb') as f:
                f.write(png)
        except IOError:
            return False
        return True

    def screenshot_as_png(self):
        return b64decode(self.screenshot_as_base64.encode('ascii'))

    def get_screenshot_as_base64(self):
        return self.screenshot_as_base64

    def get_screenshot_as_file(self, filename):
        return self.screenshot(filename)

    def get_screenshot_as_png(self):
        return self.screenshot_as_png()

    def save_screenshot(self, filename) -> bool:
        return self.screenshot(filename)



#!/usr/bin/env python3
import logging
import traceback

import click
import coloredlogs

from pymobiledevice3.cli.activation import cli as activation_cli
from pymobiledevice3.cli.afc import cli as afc_cli
from pymobiledevice3.cli.amfi import cli as amfi_cli
from pymobiledevice3.cli.apps import cli as apps_cli
from pymobiledevice3.cli.backup import cli as backup_cli
from pymobiledevice3.cli.companion_proxy import cli as companion_cli
from pymobiledevice3.cli.crash import cli as crash_cli
from pymobiledevice3.cli.developer import cli as developer_cli
from pymobiledevice3.cli.diagnostics import cli as diagnostics_cli
from pymobiledevice3.cli.usbmux import cli as usbmux_cli
from pymobiledevice3.cli.lockdown import cli as lockdown_cli
from pymobiledevice3.cli.mounter import cli as mounter_cli
from pymobiledevice3.cli.notification import cli as notification_cli
from pymobiledevice3.cli.pcap import cli as pcap_cli
from pymobiledevice3.cli.power_assertion import cli as power_assertion_cli
from pymobiledevice3.cli.processes import cli as ps_cli
from pymobiledevice3.cli.profile import cli as profile_cli
from pymobiledevice3.cli.provision import cli as provision_cli
from pymobiledevice3.cli.restore import cli as restore_cli
from pymobiledevice3.cli.springboard import cli as springboard_cli
from pymobiledevice3.cli.syslog import cli as syslog_cli
from pymobiledevice3.cli.webinspector import cli as webinspector_cli
from pymobiledevice3.exceptions import NoDeviceConnectedError, NotPairedError, UserDeniedPairingError, \
    PairingDialogResponsePendingError, SetProhibitedError, MissingValueError, DeviceHasPasscodeSetError, \
    DeveloperModeError, UsbmuxConnectionError, NoDeviceSelectedError, MessageNotSupportedError, InternalError, \
    DeveloperModeIsNotEnabledError, InvalidServiceError

coloredlogs.install(level=logging.INFO)

logging.getLogger('asyncio').disabled = True
logging.getLogger('parso.cache').disabled = True
logging.getLogger('parso.cache.pickle').disabled = True
logging.getLogger('parso.python.diff').disabled = True
logging.getLogger('humanfriendly.prompts').disabled = True
logging.getLogger('blib2to3.pgen2.driver').disabled = True
logging.getLogger('urllib3.connectionpool').disabled = True

logger = logging.getLogger(__name__)


def cli():
    cli_commands = click.CommandCollection(sources=[
        developer_cli, mounter_cli, apps_cli, profile_cli, lockdown_cli, diagnostics_cli, syslog_cli, pcap_cli,
        crash_cli, afc_cli, ps_cli, notification_cli, usbmux_cli, power_assertion_cli, springboard_cli,
        provision_cli, backup_cli, restore_cli, activation_cli, companion_cli, webinspector_cli, amfi_cli
    ])
    cli_commands.context_settings = dict(help_option_names=['-h', '--help'])
    try:
        cli_commands()
    except NoDeviceConnectedError:
        logger.error('Device is not connected')
    except ConnectionAbortedError:
        logger.error('Device was disconnected')
    except NotPairedError:
        logger.error('Device is not paired')
    except UserDeniedPairingError:
        logger.error('User refused to trust this computer')
    except PairingDialogResponsePendingError:
        logger.error('Waiting for user dialog approval')
    except SetProhibitedError:
        logger.error('lockdownd denied the access')
    except MissingValueError:
        logger.error('No such value')
    except DeviceHasPasscodeSetError:
        logger.error('Cannot enable developer-mode when passcode is set')
    except DeveloperModeError as e:
        logger.error(f'Failed to enable developer-mode. Error: {e}')
    except UsbmuxConnectionError:
        logger.error('Failed to connect to usbmuxd socket. Make sure it\'s running.')
    except MessageNotSupportedError:
        logger.error('Message not supported for this iOS version')
    except InternalError:
        logger.error('Internal Error')
    except DeveloperModeIsNotEnabledError:
        logger.error('Developer Mode is disabled. You can try to enable it using: '
                     'python3 -m pymobiledevice3 amfi enable-developer-mode')
    except InvalidServiceError:
        logger.error('Failed to access an invalid lockdown service, possibly from DeveloperDiskImage.dmg or a Cryptex. '
                     'You may try: python3 -m pymobiledevice3 mounter auto-mount')
    except NoDeviceSelectedError:
        return
    except BrokenPipeError:
        traceback.print_exc()


if __name__ == '__main__':
    cli()



import datetime
import json
import logging
import os
import uuid

import click
import coloredlogs
import inquirer
from inquirer.themes import GreenPassion
from pygments import highlight, lexers, formatters

from pymobiledevice3.exceptions import NoDeviceSelectedError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.usbmux import select_devices_by_connection_type


def default_json_encoder(obj):
    if isinstance(obj, bytes):
        return f'<{obj.hex()}>'
    if isinstance(obj, datetime.datetime):
        return str(obj)
    if isinstance(obj, uuid.UUID):
        return str(obj)
    raise TypeError()


def print_json(buf, colored=True, default=default_json_encoder):
    formatted_json = json.dumps(buf, sort_keys=True, indent=4, default=default)
    if colored:
        colorful_json = highlight(formatted_json, lexers.JsonLexer(),
                                  formatters.TerminalTrueColorFormatter(style='stata-dark'))
        print(colorful_json)
    else:
        print(formatted_json)


def set_verbosity(ctx, param, value):
    coloredlogs.set_level(logging.INFO - (value * 10))


def wait_return():
    input('> Hit RETURN to exit')


UDID_ENV_VAR = 'PYMOBILEDEVICE3_UDID'


class DeviceInfo:
    def __init__(self, lockdown_client: LockdownClient):
        self.lockdown_client = lockdown_client
        self.product_version = self.lockdown_client.product_version
        self.serial = self.lockdown_client.identifier
        self.display_name = self.lockdown_client.display_name

    def __str__(self):
        if self.display_name is None:
            return f'Unknown device, ios version: {self.product_version}, serial: {self.serial}'
        else:
            return f'{self.display_name}, ios version: {self.product_version}, serial: {self.serial}'


class Command(click.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params[:0] = [
            click.Option(('lockdown', '--udid'), envvar=UDID_ENV_VAR, callback=self.udid,
                         help=f'Device unique identifier. You may pass {UDID_ENV_VAR} environment variable to pass this'
                              f' option as well'),
            click.Option(('verbosity', '-v', '--verbose'), count=True, callback=set_verbosity, expose_value=False),
        ]

    @staticmethod
    def udid(ctx, param, value):
        if '_PYMOBILEDEVICE3_COMPLETE' in os.environ:
            # prevent lockdown connection establishment when in autocomplete mode
            return

        if value is not None:
            return LockdownClient(serial=value)

        devices = select_devices_by_connection_type(connection_type='USB')
        if len(devices) <= 1:
            return LockdownClient()

        devices_options = []
        for device in devices:
            lockdown_client = LockdownClient(serial=device.serial)
            device_info = DeviceInfo(lockdown_client)
            devices_options.append(device_info)

        device_question = [inquirer.List('device', message='choose device', choices=devices_options, carousel=True)]
        try:
            result = inquirer.prompt(device_question, theme=GreenPassion(), raise_keyboard_interrupt=True)
            return result['device'].lockdown_client
        except KeyboardInterrupt as e:
            raise NoDeviceSelectedError from e


class CommandWithoutAutopair(Command):
    @staticmethod
    def udid(ctx, param, value):
        if '_PYMOBILEDEVICE3_COMPLETE' in os.environ:
            # prevent lockdown connection establishment when in autocomplete mode
            return
        return LockdownClient(serial=value, autopair=False)


class BasedIntParamType(click.ParamType):
    name = 'based int'

    def convert(self, value, param, ctx):
        try:
            return int(value, 0)
        except ValueError:
            self.fail(f'{value!r} is not a valid int.', param, ctx)


BASED_INT = BasedIntParamType()



import click
import logging
import tempfile
from pymobiledevice3 import usbmux
from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.tcp_forwarder import TcpForwarder

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ usbmuxd cli """
    pass


@cli.group('usbmux')
def usbmux_cli():
    """ usbmuxd options """
    pass


@usbmux_cli.command('forward', cls=Command)
@click.argument('src_port', type=click.IntRange(1, 0xffff))
@click.argument('dst_port', type=click.IntRange(1, 0xffff))
@click.option('-d', '--daemonize', is_flag=True)
def usbmux_forward(lockdown: LockdownClient, src_port, dst_port, daemonize):
    """ forward tcp port """
    forwarder = TcpForwarder(lockdown, src_port, dst_port)

    if daemonize:
        try:
            from daemonize import Daemonize
        except ImportError:
            raise NotImplementedError('daemonizing is only supported on unix platforms')

        with tempfile.NamedTemporaryFile('wt') as pid_file:
            daemon = Daemonize(app=f'forwarder {src_port}->{dst_port}', pid=pid_file.name, action=forwarder.start)
            daemon.start()
    else:
        forwarder.start()


@usbmux_cli.command('list')
@click.option('--color/--no-color', default=True)
@click.option('-u', '--usb', is_flag=True, help='show only usb devices')
@click.option('-n', '--network', is_flag=True, help='show only network devices')
def usbmux_list(color, usb, network):
    """ list connected devices """
    connected_devices = []
    for device in usbmux.list_devices():
        udid = device.serial

        if usb and not device.is_usb:
            continue

        if network and not device.is_network:
            continue

        lockdown = LockdownClient(udid, autopair=False, usbmux_connection_type=device.connection_type)
        connected_devices.append(lockdown.short_info)

    print_json(connected_devices, colored=color)



#!/usr/bin/env python3

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService


class PreboardService(BaseService):
    SERVICE_NAME = 'com.apple.preboardservice_v2'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def create_stashbag(self, manifest):
        return self.service.send_recv_plist({'Command': 'CreateStashbag', 'Manifest': manifest})

    def commit(self, manifest):
        return self.service.send_recv_plist({'Command': 'CommitStashbag', 'Manifest': manifest})



import asyncio
import uuid

from fastapi import FastAPI, WebSocket
from fastapi.logger import logger

from pymobiledevice3.services.web_protocol.cdp_target import CdpTarget
from pymobiledevice3.services.web_protocol.session_protocol import SessionProtocol
from pymobiledevice3.services.webinspector import WirTypes

app = FastAPI()


@app.on_event("startup")
async def startup_event():
    app.state.inspector.connect()


@app.get('/json')
async def available_targets():
    app.state.inspector.get_open_pages()
    targets = []
    for app_id in app.state.inspector.application_pages:
        for page_id, page in app.state.inspector.application_pages[app_id].items():
            if page.type_ not in (WirTypes.WEB, WirTypes.WEB_PAGE):
                continue
            targets.append({
                'description': '',
                'id': page_id,
                'title': page.web_title,
                'type': 'page',
                'url': page.web_url,
                'webSocketDebuggerUrl': f'ws://localhost:9222/devtools/page/{page_id}',
                'devtoolsFrontendUrl': f'/devtools/inspector.html?ws://localhost:9222/devtools/page/{page_id}',
            })
    return targets


@app.get('/json/version')
def version():
    return {
        'Browser': 'Safari',
        'Protocol-Version': '1.1',
        'User-Agent': 'pymobiledevice3',
        'V8-Version': '7.2.233',
        'WebKit-Version': '537.36 (@cfede9db1d154de0468cb0538479f34c0755a0f4)',
        'webSocketDebuggerUrl': f'ws://localhost:9222/devtools/browser/{app.state.inspector.connection_id}'
    }


async def from_cdp(target: CdpTarget, websocket):
    async for message in websocket.iter_json():
        logger.debug(f'CDP INPUT:  {message}')
        await target.send(message)


async def to_cdp(target: CdpTarget, websocket):
    while True:
        message = await target.receive()
        logger.debug(f'CDP OUTPUT:  {message}')
        await websocket.send_json(message)


@app.websocket('/devtools/page/{page_id}')
async def page_debugger(websocket: WebSocket, page_id: str):
    application, page = app.state.inspector.find_page_id(page_id)
    session_id = str(uuid.uuid4()).upper()
    protocol = SessionProtocol(app.state.inspector, session_id, application, page, method_prefix='')
    target = await CdpTarget.create(protocol)
    await websocket.accept()
    await asyncio.gather(
        from_cdp(target, websocket),
        to_cdp(target, websocket),
    )



import plistlib
import typing

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService

CHUNK_SIZE = 200


class DebugServerAppList(BaseService):
    SERVICE_NAME = 'com.apple.debugserver.DVTSecureSocketProxy.applist'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def get(self) -> typing.Mapping:
        buf = b''
        while b'</plist>' not in buf:
            buf += self.service.recv(CHUNK_SIZE)

        return plistlib.loads(buf)



import logging

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import AfcService, AfcShell


class HouseArrestService(AfcService):
    SERVICE_NAME = 'com.apple.mobile.house_arrest'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        service_name = self.SERVICE_NAME
        super(HouseArrestService, self).__init__(self.lockdown, service_name)

    def send_command(self, bundle_id, cmd='VendContainer'):
        self.service.send_plist({'Command': cmd, 'Identifier': bundle_id})
        res = self.service.recv_plist()
        if res.get('Error'):
            self.logger.error('%s: %s', bundle_id, res.get('Error'))
            return False
        else:
            return True

    def shell(self, application_id, cmd='VendContainer'):
        res = self.send_command(application_id, cmd)
        if res:
            AfcShell(self.lockdown, afc_service=self).cmdloop()



from zipfile import ZipFile

import logging
import typing
from enum import Enum
from io import BytesIO

from pymobiledevice3.exceptions import NoSuchBuildIdentityError, PyMobileDevice3Exception
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.ipsw.ipsw import IPSW
from pymobiledevice3.restore.tss import TSSResponse

RESTORE_VARIANT_ERASE_INSTALL = 'Erase Install (IPSW)'
RESTORE_VARIANT_UPGRADE_INSTALL = 'Upgrade Install (IPSW)'
RESTORE_VARIANT_MACOS_RECOVERY_OS = 'macOS Customer'


class Behavior(Enum):
    Update = 'Update'
    Erase = 'Erase'


class BaseRestore:
    def __init__(self, ipsw: BytesIO, device: Device, tss: typing.Mapping = None,
                 behavior: Behavior = Behavior.Update, logger=None):
        self.logger = logging.getLogger(self.__class__.__name__) if logger is None else logger
        self.ipsw = IPSW(ZipFile(ipsw))
        self.device = device
        self.tss = TSSResponse(tss) if tss is not None else None

        if not self.device.is_image4_supported:
            raise NotImplementedError('is_image4_supported is False')

        self.logger.info(f'connected device: <ecid: {self.device.ecid} hardware_model: {self.device.hardware_model} '
                         f'image4-support: {self.device.is_image4_supported}>')

        self.logger.debug('scanning BuildManifest.plist for the correct BuildIdentity')

        variant = {
            Behavior.Update: RESTORE_VARIANT_UPGRADE_INSTALL,
            Behavior.Erase: RESTORE_VARIANT_ERASE_INSTALL,
        }[behavior]

        try:
            self.build_identity = self.ipsw.build_manifest.get_build_identity(self.device.hardware_model,
                                                                              restore_behavior=behavior.value,
                                                                              variant=variant)
        except NoSuchBuildIdentityError:
            if behavior == Behavior.Update:
                self.build_identity = self.ipsw.build_manifest.get_build_identity(self.device.hardware_model,
                                                                                  restore_behavior=behavior.value)
            else:
                raise

        self.macos_variant = None
        try:
            self.macos_variant = self.ipsw.build_manifest.get_build_identity(
                self.device.hardware_model,
                variant=RESTORE_VARIANT_MACOS_RECOVERY_OS)
            self.logger.info('Performing macOS restore')
        except NoSuchBuildIdentityError:
            pass

        build_info = self.build_identity.get('Info')
        if build_info is None:
            raise PyMobileDevice3Exception('build identity does not contain an "Info" element')

        device_class = build_info.get('DeviceClass')
        if device_class is None:
            raise PyMobileDevice3Exception('build identity does not contain an "DeviceClass" element')



import plistlib

from cached_property import cached_property

from pymobiledevice3.exceptions import NoSuchBuildIdentityError
from pymobiledevice3.restore.ipsw.build_identity import BuildIdentity


class BuildManifest:
    def __init__(self, ipsw, manifest: bytes):
        self.ipsw = ipsw
        self._manifest = plistlib.loads(manifest)
        self._parse_build_identities()

    @cached_property
    def build_major(self):
        build_major = str()
        for i in self._manifest['ProductBuildVersion']:
            if i.isdigit():
                build_major += i
            else:
                break

        return int(build_major)

    @cached_property
    def supported_product_types(self):
        return self._manifest['SupportedProductTypes']

    @cached_property
    def supported_product_types_family(self):
        product = self.supported_product_types[0]
        if product.startswith('iBridge'):
            return 'iBridge'
        elif product.startswith('iPhone'):
            return 'iPhone'
        else:
            raise ValueError()

    def get_build_identity(self, device_class: str, restore_behavior: str = None, variant: str = None):
        for build_identity in self.build_identities:
            if variant is not None:
                if variant not in build_identity.variant:
                    continue

            if build_identity.device_class != device_class:
                continue

            if restore_behavior is not None:
                if build_identity.restore_behavior != restore_behavior:
                    continue

            return build_identity
        raise NoSuchBuildIdentityError('failed to find the correct BuildIdentity from the BuildManifest')

    def _parse_build_identities(self):
        self.build_identities = []
        for build_identity in self._manifest['BuildIdentities']:
            self.build_identities.append(BuildIdentity(self, build_identity))



from datetime import datetime, timedelta

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import load_pem_public_key, Encoding, PrivateFormat, NoEncryption


def make_cert(key, public_key, common_name=None):
    attributes = [x509.NameAttribute(NameOID.COMMON_NAME, common_name)] if common_name else []
    subject = issuer = x509.Name(attributes)
    cert = x509.CertificateBuilder()
    cert = cert.subject_name(subject)
    cert = cert.issuer_name(issuer)
    cert = cert.public_key(public_key)
    cert = cert.serial_number(1)
    cert = cert.not_valid_before(datetime.utcnow() - timedelta(minutes=1))
    cert = cert.not_valid_after(datetime.utcnow() + timedelta(days=365 * 10))
    cert = cert.sign(key, hashes.SHA1())
    return cert


def dump_cert(cert):
    return cert.public_bytes(Encoding.PEM)


def ca_do_everything(device_public_key):
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    cert = make_cert(private_key, private_key.public_key())
    dev_key = load_pem_public_key(device_public_key)
    dev_cert = make_cert(private_key, dev_key, 'Device')
    return dump_cert(cert), private_key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()), dump_cert(
        dev_cert)
